



type point        float x,y
type pointcart    float x,y
type pointpolar   float r,a
type line         point p1,p2
type boundingrect point p1,p2


'2D GRAPHICS SUPPORT
'===================

sub drawline(sys q, n=1)
========================
float p at (q)
'float*p=q
sys i
glBegin GL_LINE_STRIP
glVertex2fv p
for i=1 to n
  @p+=8 : glVertex2fv p
next 
glEnd
end sub

sub drawpoint(sys q, n=1)
=========================
float p at (q)
sys i
glBegin GL_POINTS
for i=1 to n
  glVertex2fv p : @p+=8
next
glEnd
end sub


=================================


include once "ParseUtil.inc"

sub StringToFloat(string s, sys*i, float*f, sys *n)
===================================================
sys p=strptr s
do
  stepword s,i
  if not ascw then exit do
  f=valat p+sttw-1
  n++ : @f+=4
end do
end sub

sub FloatToString(string s, sys*i, float*f, sys *n)
===================================================
sys    j
string wr
indexbase 1
for j=1 to n
  wr=str f
  mid s,i,wr : i+=len wr
  mid s,i," " : i++ 
  @f+=4
next
end sub


=================================



function samepos(point *p,*q) as sys
====================================
if p.x = q.x and p.y = q.y then return 1
end function

function dist(point *a,*b) as float
===================================
return hypot(b.x-a.x , b.y-a.y)
end function


function interpoint(point *a,*b,*p, float f)
============================================
p.x=a.x+(b.x-a.x)*f
p.y=a.y+(b.y-a.y)*f
end function

function midpoint(point *a,*b,*p)
=================================
static float half=0.5
p.x=a.x+(b.x-a.x)*half
p.y=a.y+(b.y-a.y)*half
end function

function rightangle(point *pa,*pb,*pc,*pd) as sys
=================================================
'anticlockwise triangle points:
'left  pa
'right pb
'apex  pc
'perpendicular between pc pd, forming 2 triangles
float a,b,c,d
a=dist pc,pa 'LEFT SIDE
b=dist pc,pb 'RIGHT SIDE
c=dist pa,pb 'BASE
if c<>0 then
  d=(a*a-b*b+c*c)/(2*c)
  interpoint pa,pb,pd, d/c
  return 1
end if
end function

function calcgradient(line *l, float *g,*i)
===========================================
float x = l.p2.x - l.p1.x
float y = l.p2.y - l.p1.y
if x=0 then
  g=0 : i = 1 'SET VERTICAL LINE FLAG
else
  g = y/x : i=0
end if
end function

function rot90(point *p,*q,*r)
==============================
float ddx,ddy
ddx = q.x-p.x
ddy = q.y-p.y
r.x = p.x-ddy
r.y = p.y+ddx
end function

function rotm90(point *p,*q,*r)
===============================
float ddx,ddy
ddx = q.x-p.x
ddy = q.y-p.y
r.x = p.x+ddy
r.y = p.y-ddx
end function

function topolar(point *p,*q,*r) as sys
=======================================
float dx,dy
dx=q.x-p.x
dy=q.y-p.y
r.x=hypot(dx,dy) 'as polar dist
r.y=atan(dy,dx)  'as polar angle
return 1
end function

function tocart(point *p,*q,*r) as sys
======================================
float dx,dy
dx=r.x*cos(r.y) 'r.x polar dist
dy=r.x*sin(r.y) 'r.y polar angle
q.x+=dx
q.y+=dy
return 1
end function

function inside(point *a,*b,*c,*p) as sys
=========================================
float ab,bc,ca,pa,pb,pc
ab=dist a,b
bc=dist b,c
ca=dist c,a
pa=dist p,a
pb=dist p,b
pc=dist p,c
if samepos(a,p) then return 1
if samepos(b,p) then return 2
if samepos(c,p) then return 4
if ca+bc > pa+pb and ab+ca > pb+pc and bc+ab > pc+pa then return 16
end function

function exclusive(line *l1,*l2) as sys
=======================================
float x,y
boundingrect b1,b2
copy @b1,@l1,sizeof boundingrect
copy @b2,@l2,sizeof boundingrect
if b1.p1.x>b1.p2.x then x=b1.p1.x : b1.p1.x=b1.p2.x : b1.p2.x=x
if b1.p1.y>b1.p2.y then y=b1.p1.y : b1.p1.y=b1.p2.y : b1.p2.y=y
if b2.p1.x>b2.p2.x then x=b2.p1.x : b2.p1.x=b2.p2.x : b2.p2.x=x
if b2.p1.y>b2.p2.y then y=b2.p1.y : b2.p1.y=b2.p2.y : b2.p2.y=y
if b1.p2.x<b2.p1.x then return 1
if b1.p1.x>b2.p2.x then return 1
if b1.p2.y<b2.p1.y then return 1
if b1.p1.y>b2.p2.y then return 1
end function

function intersect(line *l1,*l2, point *p) as sys
=================================================
' 0 no intersection
' 1 same pos as l1.p1
' 2 same pos as l1.p2
' 4 same pos as l2.p1
' 8 same pos as l2.p2
'16 intersection: lines cross between points
'
float g1,g2,i1,i2
calcgradient l1,g1,i1
calcgradient l2,g2,i2
if g1=g2 and i1=i2 then
  return 0 'PARALLEL LINES
  'itr overlapping lines ?
end if
float xd,yd,y2v,xp,yp
xd=l2.p1.x - l1.p1.x
if i2 then
  'vertical line l2
  xp=l2.p1.x
  yp=l1.p1.y+g1*xd
elseif i1 then
  'vertical line l1
  xp=l1.p1.x
  yp=l2.p1.y-g2*xd
else
'print g1 "  " g2
  y2v = l2.p1.y - xd*g2
  yd  = y2v - l1.p1.y
  xp  = yd/(g1-g2)
  yp  = l1.p1.y + xp*g1
  xp += l1.p1.x
end if
p = {xp,yp}
return 1
end function

function boundstatus(line *l1,*l2, point *p) as sys
===================================================
sys   a=0
float xp=p.x , yp=p.y
'
'COINCIDING POINT TESTING
'
if samepos (p,l1.p1) then a or=1
if samepos (p,l1.p2) then a or=2
if samepos (p,l2.p1) then a or=4
if samepos (p,l2.p2) then a or=8

if a then return a

'LINE LIMIT TESTING
'
if l1.p1.x=l1.p2.x then
  if xp=l1.p1.x then a=16
elseif xp>l1.p1.x and xp<l1.p2.x then
  a=16
elseif xp>l1.p2.x and xp<l1.p1.x then
  a=16
end if
'
if a then a=0 else return 0
'
if l2.p1.x=l2.p2.x then
  if xp=l2.p1.x then a=16
elseif xp>l2.p1.x and xp<l2.p2.x then
  a=16
elseif xp>l2.p2.x and xp<l2.p1.x then
  a=16
end if
'
if a then a=0 else return 0
'
if l1.p1.y=l1.p2.y then
  if yp=l1.p1.y then a=16
elseif yp>l1.p1.y and yp<l1.p2.y then
  a=16
elseif yp>l1.p2.y and yp<l1.p1.y then
  a=16
end if
'
if a then a=0 else return 0
'
if l2.p1.y=l2.p2.y then
  if yp=l2.p1.y then a=16
elseif yp>l2.p1.y and yp<l2.p2.y then
  a=16
elseif yp>l2.p2.y and yp<l2.p1.y then
  a=16
end if
return a
end function

function intersected(line *l1,*l2,*point p) as sys
==================================================
if exclusive(l1,l2) then
  return 0
elseif intersect(l1,l2,p) then
  return boundstatus l1,l2,p
end if
end function

function centrefrom3points(point *p,*q,*r,*c) as sys
====================================================
point mpq,mqr
point npq,nqr
sys   a
midpoint p,q,mpq
midpoint q,r,mqr
rotm90 mpq,q,npq 
rotm90 mqr,r,nqr
line l1,l2
l1.p1.x = mpq.x : l1.p1.y=mpq.y
l1.p2.x = npq.x : l1.p2.y=npq.y
l2.p1.x = mqr.x : l2.p1.y=mqr.y
l2.p2.x = nqr.x : l2.p2.y=nqr.y
a=intersect l1,l2,c
return a
end function

function arcfrom3points(point *p,*q,*r,*c,*s,*e) as sys
=======================================================
sys   a
a=centrefrom3points p,q,r,c
if a then
  topolar c,p,s
  topolar c,r,e
end if
return @r
end function


