def useLispishStorage {
=====================

'single buffer for embedded names and contents

string varbuf  'store vars and funs
sys    varend  'length of variable buffer content
sys    bufend  'length of variable buffer

%% StretchUnit 0x10000

function stretchbuffer(sys le)
==============================
if varend+le>bufend then
  varbuf+=nuls(stretchunit+le) : bufend+=stretchunit+le
end if
end function


function findvar(string wr, sys *i,*e)
======================================
'SEARCHING BACKWARDS
byte w at strptr wr
byte b at strptr varbuf
sys lw,ld,ln,bg,j
lw=len wr
bg=@b
@b+=varend
do
  @b-=4       'point to data length
  if @b<=bg then
    i=0 : e=0
    return 'not found
  end if
  ld=(long) b    'length of data (excl null)
  @b=@b-ld-2     'skip to name length
  ln=b           'length of name
  @b=@b-ln       'point to first char of name
  if ln<>lw then continue do 'mismatched lengths
  '
  'BYTE MATCHING
  j=1
  do
    bt=b[j]
    'COMPLETE MATCH
    if j>lw then
      i=@b-bg+ln+2 'START OF DATA
      e=i+ld       'END BOUNDARY OF DATA
      return
    end if
    if bt<>w[j] then exit do 'NO MATCH
    j++
  end do
end do
end function


macro setNewVar
===============
sys ee=lw+le+6                   'format |word:lw:data:0:le|
stretchbuffer ee
sys p=varend+strptr varbuf
copy p,strptr(wr),lw     : p+=lw 'name
(byte) *p=lw             : p++   'name length
copy p,i+strptr(sr),le   : p+=le 'data
(byte) *p=0              : p++   'terminating null byte
(long) *p=le             :       'datalength (incl null)
varend+=ee
end macro


macro setVirtualTrim
====================
'EFFECTIVE LEFT TRIM
i=1 : skipspace(sr,i) : i--
'EFFECTIVE RIGHT TRIM
byte b at le+strptr sr
le-=i
do
  if le<=0 then exit do
  @b--
  if b>32 then exit do
  le--
end do
end macro


function newvar(string wr,sr)
=============================
sys i
sys lw=len wr
sys le=len sr
setNewVar
end function


function newtrimvar(string wr,sr)
=================================
sys i
sys lw=len wr
sys le=len(sr)
setVirtualTrim 
setNewVar
end function


function insvar(string s, sys aa,bb)
====================================
sys ls=len s
sys lv=bb-aa 'length of gap available
sys j=ls-lv  'gap contracts or expands
sys i,k
if j>0 then stretchbuffer j
byte b at strptr varbuf
if j<0 'CONTRACT -ve j
  k=bb+j
  for i=bb to varend : b[k]=b[i] : k++ : next
elseif j>0 'EXPAND +j
  k=varend+j
  for i=varend to bb step -1 : b[k]=b[i] : k-- : next
end if  
varend+=j
sys p=@b+aa-1
copy p,strptr(s),ls
p+=ls          'skip data
(byte) *p=0    'null byte
p++            'data length field
(long) *p=ls   'data length excl null
end function

macro newScope(v) {v=varend}
macro endScope(v) {varend=v}

} 'end def


def UseStorageClasses {
=====================

type linkedunit
  sys last
  sys typ
  sys hash
  string nam
  union
    string  ss
    sys     ii
    float   ff
    double  dd
    quad    qq
  end union
end type

'#recordof linkedunit


class linkedlist
================
  '
  string bu
  linkedunit *u, *uu
  sys max
  sys pos
  def su  {sizeof(linkedunit)}
  '
  method last() as sys
    sys p=pos/su
    return p-1
  end method
  '
  method add(string n,s, sys *p) as sys
    if pos+su > max then 'buffering
      bu+=nuls 0x4000 : max+=0x4000
      @u=strptr bu
    end if
    @uu=@u+pos
    uu.last=p
    uu.typ=1
    uu.nam=n
    uu.ss=s
    p=pos/su
    pos+=su
    return -1
  end method
  '
  method find(string f, int*p) as sys
  if p<0 then return 0  
  do
    @uu=p*su+@u
    if f=uu.nam then return -1
    p=uu.last
    if p<0 then exit do 'otherwise check previous
  end do
  end method
  '
  method out(string n,*s, int*f) as sys
    if find n,f then s=uu.ss : return -1
    s=""
  end method
  '
  method in(string n,s, int *f) as sys
    if find n,f then uu.ss=s : uu.typ=1 : return -1
  end method
  '
end class


class scopedStorage
===================
  '
  has linkedList li
  int sc[0x4000] 'supports 64 scope levels each 256 ints (1024 bytes) ascii-based entries
  int scopeoffset
  '
  macro locator(p) {int p at @sc+asc(n,1)*4+scopeoffset}
  '
  method init()
  int *p=@sc
  int i
  for i=1 to 0x4000
    p=-1
    @p+=4
  next
  end method
  '
  method add(string n,s) as sys
    locator p
    li.add n,s,p 'new link p is now set in the sc table
  end method
  '
  method find(string n, int *f) as sys
    locator p
    f=p
    return li.find n,f
  end method
  '
  method in(string n,*s) as sys
    locator p
    int q=p
    return li.in n,s,q
  end method
  '
  method out(string n,*s) as sys
    locator p
    int q=p
    return li.out n,s,q
  end method
  '
  method newScope()
    int *p,*q
    @p=@sc+scopeoffset
    p=li.pos
    @q=@p+0x400
    copy @q,@p,0x400
    scopeoffset+=0x400
  end method
  '
  method endScope()
    if scopeoffset then scopeoffset-=0x400
    int p at @sc+scopeoffset
    li.pos=p
  end method
  '
end class

} 'end def